
// Copyright 2009 - Christopher Ramrez
// This unit had been realesed under the public domain.
// LICENSE:
//   You must release under public domain all modifications
//   and/or enhancements you make to this unit.

        /////////////////////////////////////////////////////
        //                                                 //
        //          >>>>> PLEASE NOTE <<<<<<               //
        //  THIS UTILITY DOES NOT SUPPORT "ROLES"          //
        //  IT'S UP TO YOU TO WRITE THIS SUPPORT           //
        //                    <<>>                         //
        //  ESTA UNIDAD NO SOPORTA LOS "ROLES"             //
        //  ES CUESTIN SUYA ESCRIBIR ESTA FUNCIONALIDAD   //
        //                                                 //
        /////////////////////////////////////////////////////
        
        
        // THIS FUNCTION REQUIRES:
        // Unified Interbase components (UIB) -> These components are libre
        
        
        // EXAMPLE USE:
        // 
        // Test if curent user user has insert privilege over 'ORDERS' table:
        //   if not PrivilegeQueryInitializated then
        //     if not FetchPrivileges(current_connect_user) then
        //        raise_an_exception;
        //
        //   InsertAction.Enabled := UserHasPrivilege('ORDERS', fbpInsert);
        

unit fb_credentials_parser;

interface

  uses
    UIB, SysUtils, StrUtils;

    CONST
        // consts for original structure of RDB$USER_PRIVILEGES
        // system table in a normal firebird 2.1 server
        USER_FIELD          = 'RDB$USER';
        GRANTOR_FIELD       = 'RDB$GRANTOR';
        PRIVILEGE_FIELD     = 'RDB$PRIVILEGE';
        GRANT_OPTION_FIELD  = 'RDB$GRANT_OPTION';
        RELATION_FIELD      = 'RDB$RELATION_NAME';  // stores table name
        FIELDNAME_FIELD     = 'RDB$FIELD_NAME';
        
        
        // Structure for an enhanced privileges query
        ehUSER_FIELD        = 'RDB$USER';
        ehCRUD_FIELD        = 'RDB$CRUD';
        ehGRAND_FIELD       = 'RDB$GRAND_OPTION';
        ehRELATION_FIELD    = 'RDB$RELATION_NAME';
        ehFIELDS_NAME       = 'RDB$FIELD_NAME';
        
        
        // the following is the SQL query necesary to get an ehanced
        // crud privilege query. This query requires FB 2.1 or above
        EHANCED_CRUD_QUERY =
            'SELECT '                                                   +
                'a.RDB$USER,'                                           +
                'LIST(trim(a.RDB$PRIVILEGE)) AS RDB$CRUD,'              +
                'LIST(a.RDB$GRANT_OPTION) AS RDB$GRANT_OPTION,'         +
                'a.RDB$RELATION_NAME,'                                  +
                'a.RDB$FIELD_NAME '                                     +
            'FROM '                                                     +
                '('                                                     +
                 'SELECT '                                              +
                    'UP.RDB$USER,'                                      +
                    'UP.RDB$RELATION_NAME,'                             +
                    'UP.RDB$GRANT_OPTION,'                              +
                    'UP.RDB$PRIVILEGE,'                                 +
                    'LIST(TRIM(UP.RDB$FIELD_NAME)) as RDB$FIELD_NAME '  +
                 'FROM RDB$USER_PRIVILEGES UP '                         +
                 'WHERE UPPER(UP.RDB$USER) in (''%s'', ''PUBLIC'') '    +
                 'GROUP BY UP.RDB$PRIVILEGE,'                           +
                          'UP.RDB$GRANT_OPTION,'                        +
                          'UP.RDB$USER,'                                +
                          'UP.RDB$RELATION_NAME'                        +
                ') as a '                                               +
            'GROUP BY '                                                 +
                'a.RDB$USER,'                                           +
                'a.RDB$RELATION_NAME,'                                  +
                'a.RDB$FIELD_NAME';
    
    type
        TbioFBPrivilege = (fbpSelect, fbpUpdate, fbpDelete, fbpInsert, fbpReference, fbpExecute);
        PbioFBPrivilege = ^TbioFBPrivilege;
        TbioFBPrivileges = set of TbioFBPrivilege;
        PbioFBPrivileges = ^TbioFBPrivileges;

        TbioFBPrivilegeEx = record
          FBPrivile: TbioFBPrivilege;
          // ApplyOnFields apply only when FBPrivilege is fbpUpdate or fbpReference
          ApplyOnFields: String;
        end;
        PbioFBPrivilegeEx = ^TbioFBPrivilegeEx;

        TbioFBPrivilegeExs = array of TbioFBPrivilegeEx;
        PbioFBPrivilegeExs = ^TbioFBPrivilegeExs;

        
    //  published functions in this unit
    function GetUserPrivilegesOnTable(RelationName: String): TbioFBPrivileges;
    function UserHasPrivilege(RelationName: String;
                              Privilege: TbioFBPrivilege): Boolean;
    function UserHasGrantOption(RelationName: String;
                                Privilege: TbioFBPrivilege): Boolean;
    function GetPrivilegesWithGrantOption(RelationName: String): TbioFBPrivileges;
    function GetPrivilegeOnlyOnFields(RelationName: String;
                                      Privilege: TbioFBPrivilege): String;
                                      
    function FetchPrivileges(Username: String): Boolean; // REIMPLEMENT THIS FUNCTION FOR YOUR NEEDS
    function PrivilegeQueryInitializated(TestForActivatedQuery: Boolean = True): Boolean;
    function ReleasePrivilegesQuery: Boolean;
    procedure RaiseCouldNotConsultPrivileges;

    // The following two functions are for specials cases and only should be
    // used if you understan how they work
    function ReFetchPrivileges(Username: String): Boolean; // REIMPLEMENT THIS FUNCTION FOR YOUR NEEDS
    function GetPrivilegeQuery: TuibQuery;

implementation

  uses bioDataModule_imp;


 (*
 | RDB$USER | RDB$GRUD  | RDB$GRANT_OPTION | RDB$RELATION_NAME | RDB$FIELD_NAME               |
 | -------------------------------------------------------------------------------------------|
 | CHRIS    | S,I,D     | 1,0,0            | NOMBRE_TABLA_1    | [nul]                        |
 | CHRIS    | U         | 0                | NOMBRE_TABLA_1    | NOMBRES,APELLIDOS,OTRO_CAMPO |
 | CHRIS    | S         | 0                | NOMBRE_TABLA_2    | [nul]                        |
 *)

    var
    FPrivilegesQuery: TuibQuery; // this object stores the privileges taken from
                                 // the server. Must be initializated with a
                                 // FetchPrivileges function call
                                 
procedure raise_PrivilegesQuery_not_initializated;
begin

    Exception.Create('No se ha iniciado an el objeto FPrivilegesQuery. Hagalo haciendo una llamada a la funcin FetchPrivileges.');
    
end;

function GetUserPrivilegesOnTable(RelationName: String): TbioFBPrivileges;
begin

    // This function locates all records with RelationName in ehTABLE_FIELD
    // field and Username param in ehUSER_FIELD. Then, enumerates all the
    // GRUD members in ehGRUB_FIELD and returns a TbioFBPrivileges based
    // on the found members.
    
    Result := [];
    
    if UserHasPrivilege(RelationName, fbpSelect) then
        Result := ( Result + [fbpSelect] );
        
    if UserHasPrivilege(RelationName, fbpInsert) then
        Result := ( Result + [fbpInsert] );
    
    if UserHasPrivilege(RelationName, fbpDelete) then
        Result := ( Result + [fbpDelete] );
    
    if UserHasPrivilege(RelationName, fbpUpdate) then
        Result := ( Result + [fbpUpdate] );
    
    if UserHasPrivilege(RelationName, fbpReference) then
        Result := ( Result + [fbpReference] );
        
    (*
    with FPrivilegesQuery then
    begin
    
        Result := [];
        
        if not PrivilegeQueryInitializated then
            raise_PrivilegesQuery_not_initializated;
            
        First;
        
        while not LocateNext(vararraycreate([ehRELATION_FIELD]),
                             vararraycreate([RelationName]),
                             [loCaseInsensitive]) do
        begin
        
            if Pos('S', Fields[ehCRUB_FIELD].AsString) > 0 then
                Result := Result + fbpSelect;
                
            if Pos('I', Fields[ehCRUB_FIELD].AsString) > 0 then
                Result := Result + fbpInsert;
                
            if Pos('D', Fields[ehCRUB_FIELD].AsString) > 0 then
                Result := Result + fbpDetele;
                
            if Pos('U', Fields[ehCRUB_FIELD].AsString) > 0 then
                Result := Result + fbpUpdate;
                
            if Pos('R', Fields[ehCRUB_FIELD].AsString) > 0 then
                Result := Result + fbpReference;
        
        end;
    
    end;
    *)
end;


function UserHasPrivilege(RelationName: String;
                          Privilege: TbioFBPrivilege): Boolean;
begin

    // this function returns true is the user has a specific Privilege
    // over the item indicated in RelationName

    Result := False;
    
    with FPrivilegesQuery do
    begin
        
        if not PrivilegeQueryInitializated then
            raise_PrivilegesQuery_not_initializated;
            
        First;

        while not {LocateNext(vararraycreate([ehRELATION_FIELD]),
                             vararraycreate([RelationName]),
                             [loCaseInsensitive])} Eof do
        begin

            if not AnsiSameText(RelationName, trim(Fields.ByNameAsAnsiString[ehRELATION_FIELD])) then
            begin
                Next;
                Continue;
            end;

            case Privilege of
            
                fbpSelect:
                    Result := (Pos('S', Fields.ByNameAsUnicodeString[ehCRUD_FIELD]) > 0);
                    
                fbpInsert:
                    Result := (Pos('I', Fields.ByNameAsUnicodeString[ehCRUD_FIELD]) > 0);

                fbpDelete:
                    Result := (Pos('D', Fields.ByNameAsUnicodeString[ehCRUD_FIELD]) > 0);
                    
                fbpUpdate:
                    Result := (Pos('U', Fields.ByNameAsUnicodeString[ehCRUD_FIELD]) > 0);
                    
                fbpReference:
                    Result := (Pos('R', Fields.ByNameAsUnicodeString[ehCRUD_FIELD]) > 0);

                fbpExecute:
                    Result := (Pos('X', Fields.ByNameAsUnicodeString[ehCRUD_FIELD]) > 0);

            end;
            
            if Result then
                Break;

            Next;
        
        end;
    
    end;
    
end;

function UserHasGrantOption(RelationName: String;
                            Privilege: TbioFBPrivilege): Boolean;
begin

    // Returns true is the user has a Privilege with Grant authorization
    
    if UserHasPrivilege(RelationName, Privilege) then
    with FPrivilegesQuery do
    begin
 
        case Privilege of
        
            fbpSelect:
                Result := (Fields.ByNameAsUnicodeString[ehGRAND_FIELD][Pos('S', Fields.ByNameAsUnicodeString[ehCRUD_FIELD])] = '1');
                
            fbpInsert:
                Result := (Fields.ByNameAsUnicodeString[ehGRAND_FIELD][Pos('I', Fields.ByNameAsUnicodeString[ehCRUD_FIELD])] = '1');
                
            fbpDelete:
                Result := (Fields.ByNameAsUnicodeString[ehGRAND_FIELD][Pos('D', Fields.ByNameAsUnicodeString[ehCRUD_FIELD])] = '1');
                
            fbpUpdate:
                Result := (Fields.ByNameAsUnicodeString[ehGRAND_FIELD][Pos('U', Fields.ByNameAsUnicodeString[ehCRUD_FIELD])] = '1');
                
            fbpReference:
                Result := (Fields.ByNameAsUnicodeString[ehGRAND_FIELD][Pos('R', Fields.ByNameAsUnicodeString[ehCRUD_FIELD])] = '1');
        
        end; 
    
    end;
    
end;

function GetPrivilegesWithGrantOption(RelationName: String): TbioFBPrivileges;
begin

    // returns an enumeration indicating all user's privileges with
    // grant authorization

    Result := [];
    
    if UserHasGrantOption(RelationName, fbpSelect) then
        Result := ( Result + [fbpSelect] );
        
    if UserHasGrantOption(RelationName, fbpInsert) then
        Result := ( Result + [fbpInsert] );
    
    if UserHasGrantOption(RelationName, fbpDelete) then
        Result := ( Result + [fbpDelete] );
    
    if UserHasGrantOption(RelationName, fbpUpdate) then
        Result := ( Result + [fbpUpdate] );
    
    if UserHasGrantOption(RelationName, fbpReference) then
        Result := ( Result + [fbpReference] );
        
end;

function GetPrivilegeOnlyOnFields(RelationName: String;
                                  Privilege: TbioFBPrivilege): String;
begin

    // Returns the value of ehFIELDS_NAME field.
    // NOTE: As this function may return a null value, indicating either that 
    // the user may has privilege on all fields, or as well that he/she
    // doesn't have Privilege in RelationName. To avoid this confusion,
    // you should use UserHasPrivilege function along with this one.

    Result := '';
    
    if UserHasPrivilege(RelationName, Privilege) then
    with FPrivilegesQuery do
    begin
    
        Result := Fields.ByNameAsUnicodeString[ehFIELDS_NAME];

    end;
    
end;

                                  
function FetchPrivileges(Username: String): Boolean;
begin

    // ***** REIMPLEMENT THIS FUNCTION FOR YOUR NEEDS *****
    // this function creates object FPrivilegesQuery and fetch the
    // EHANCED_CRUD_QUERY query results for use in this unit.

    if PrivilegeQueryInitializated then
        Exit;
        
    FPrivilegesQuery              := TuibQuery.Create(nil);
    FPrivilegesQuery.SQL.Text     := Format(EHANCED_CRUD_QUERY, [UpperCase(Username)]);
    
    // STABLISH YOUR APPLICATION DEFAULT UIB DATABASE AND TRANSACTION
    FPrivilegesQuery.DataBase     := my_application_datamodule.default_uib_database;
    FPrivilegesQuery.Transaction  := my_application_datamodule.default_uib_transaction;
    
    FPrivilegesQuery.FetchBlobs   := True;
    FPrivilegesQuery.Open;

    Result := True;

end;

function PrivilegeQueryInitializated(TestForActivatedQuery: Boolean = True): Boolean;
begin

    Result := (FPrivilegesQuery <> nil);

    if (Result and TestForActivatedQuery) then
    begin
        FPrivilegesQuery.Open(True);
        Result := True;
    end;

end;

function ReleasePrivilegesQuery: Boolean;
begin

    if PrivilegeQueryInitializated then
    begin
        FPrivilegesQuery.Close;
        FPrivilegesQuery.Free;
    end;
    
end;

procedure RaiseCouldNotConsultPrivileges;
resourcestring
  EXCEPTION_MSG =
    'No se pudieron consultar los privilegos que usted tiene.';
begin

  Exception.Create(EXCEPTION_MSG);

end;

// ****************** SPECIAL FUNCTIONS ****************** //
function ReFetchPrivileges(Username: String): Boolean;
begin

    // ***** REIMPLEMENT THIS FUNCTION FOR YOUR NEEDS *****
    // this function creates object FPrivilegesQuery and ReFetch the
    // EHANCED_CRUD_QUERY query results for use in this unit.
    //
    // This function is intended to fetch a different user's privileges.
    // A different user of the one already fetched with FetchPrivileges.

    if not PrivilegeQueryInitializated(False) then
      FPrivilegesQuery := TuibQuery.Create(nil)
    else
      FPrivilegesQuery.Close();

    FPrivilegesQuery.SQL.Text     := Format(EHANCED_CRUD_QUERY, [UpperCase(Username)]);
    
    // STABLISH YOUR APPLICATION DEFAULT UIB DATABASE AND TRANSACTION
    FPrivilegesQuery.DataBase     := my_application_datamodule.default_uib_database;
    FPrivilegesQuery.Transaction  := my_application_datamodule.default_uib_transaction;
    
    FPrivilegesQuery.FetchBlobs   := True;
    FPrivilegesQuery.Open;

    Result := True;

end;

function GetPrivilegeQuery: TuibQuery;
begin

    Result := FPrivilegesQuery;

end;


end.

    finalization
        if PrivilegeQueryInitializated(false) then
            ReleasePrivilegesQuery;